#!/usr/bin/env python3
"""Python library and CLI to search in multiple torrent sources.

Uses torrentmirror to get the first open mirror for any of the
supported sites.

Currently supported sites:

- ThePirateBay
- KickAssTorrents
"""

from contextlib import suppress
import abc
import itertools
import re
import subprocess

from blessings import Terminal
from docopt import docopt
from inquirer import List
from inquirer import prompt
from pygogo import Gogo

from requests import Session
import requests

import robobrowser
import tableprint
import torrentmirror


class BaseSearch(metaclass=abc.ABCMeta):
    """Base Search."""

    def __init__(self, logger):
        """Initialize browser instance."""
        session = Session()
        session.verify = False
        self.logger = logger
        self.browser = robobrowser.RoboBrowser(
            session=session, parser='html.parser', timeout=50)

    @abc.abstractproperty
    def url(self):
        """Search on given url by default."""

    @abc.abstractproperty
    def proxy_name(self):
        """Proxy name to use from torrentmirror."""

    @abc.abstractproperty
    def url_format(self):
        """Url formatter to add to base_url."""

    @abc.abstractmethod
    def get_torrents(self):
        """Return a list of torrents."""

    def search_magnets(self, query: str, page: int):
        """Return the list of magnets from a specific page."""
        proxies = torrentmirror.get_proxies()[self.proxy_name]
        proxies.insert(0, [self.url, None])
        for site, _ in proxies:
            self.logger.debug("Searching in %s", site)
            with suppress(requests.exceptions.ReadTimeout, AssertionError):
                http = '' if site.startswith('http') else 'http://'
                self.browser.open(self.url_format.format(
                    http, site, query, page))
                torrents = self.get_torrents()
                assert torrents
                self.logger.debug("Got torrents %s", torrents)
                return torrents

    def search(self, query: str, pagelen: int = 1):
        """Manage complete search for specified pages."""
        return itertools.chain(*(self.search_magnets(query, page + 1)
                                 for page in range(0, pagelen)))


class ThePirateBay(BaseSearch):
    """ThePirateBay search engine.

    Search on any of the pirate bay proxies or thepiratebay itself if
    available. Extract torrents, description and return it tabulated and ready
    to be printed
    """

    proxy_name = 'The Pirate Bay'
    url_format = '{}{}/search/{}/{}/99'
    url = 'https://thepiratebay.org/'

    @staticmethod
    def tabulate(link):
        """Make a given a href link parsed on bs4 printable."""
        return (link.parent.find(class_='detName').text,
                link.parent.find(class_='detDesc').text,
                link['href'])

    def get_torrents(self):
        """Return a list of torrents, printable."""
        # pylint: disable=not-callable
        torrents = self.browser.find_all(
            'a', title='Download this torrent using magnet')
        return [self.tabulate(torrent) for torrent in torrents]


class Katcr(BaseSearch):
    """KickAssTorrents search engine.

    Search on any of the pirate bay proxies or thepiratebay itself if
    available. Extract torrents, description and return it tabulated and ready
    to be printed
    """

    proxy_name = 'Kickass Torrents'
    url_format = '{}{}/search.php?q={}&p={}'
    url = 'https://kat.cd'

    @staticmethod
    def tabulate(link):
        """Make a given a href link parsed on bs4 printable."""
        magnet = link.find('a', title='Torrent magnet link')
        return (link.find(class_='cellMainLink').text,
                link.find_all('td')[1].text,
                magnet['href'])

    def get_torrents(self):
        """Return a list of torrents, printable."""
        # pylint: disable=not-callable
        torrents = self.browser.find_all(id=re.compile(
            'torrent_latest_torrents(.*)'))
        return [self.tabulate(torrent) for torrent in torrents]


def get_short(where, what):
    """Get magnet short link."""
    return '{}/{}'.format(
        where, requests.post(where, data={'magnet': what}).text)


def get_from_short(shortener, search_res):
    """Get new search res from shortened urls."""
    for elem in search_res:
        yield elem[:-1] + (get_short(shortener, elem[-1]),)


def limit_terminal_size(what, limit=-20):
    """Limit a string to current terminal size, plus limit."""
    return what[:Terminal().width + limit]


def main():
    """Search in multiple torrent sites.

    Usage: katcr [options] [interactive options] <SEARCH_TERM>

    Currently available search engines:

    - Katcr
    - ThePirateBay

    Options:
        -e --search-engine=<SearchEngine>  Torrent search engine to use
                                           [default: Katcr].
        -p --pages=<PAGES_NUM>             Number of pages to lookup
                                           [default: 1]
        -d --disable-shortener             Disable url shortener
        -s --shortener=<SHORTENER_URL>     Use given magnet shortener to
                                           prettify urls.
                                           [default: http://www.shortmag.net]

    Interactive Options:
        -i --interactive                   Enable interactive mode
        -o --open                          Launch with default torrent app
                                           in interactive mode [default: True]
        -h --help                          Show this help screen
        -v --verbose                       Enable debug mode
    """
    opt = docopt(main.__doc__, version="0.0.1")
    logger = Gogo(__name__, verbose=opt.get('--verbose')).logger

    search_res = list(globals()[opt['--search-engine'][0]](logger).search(
        opt["<SEARCH_TERM>"], int(opt.get("--pages")[0])))

    if not opt['--disable-shortener']:
        search_res = list(get_from_short(opt['--shortener'][0], search_res))

    if not search_res:
        return

    if not opt['--interactive']:
        return tableprint.table(search_res, ['Description', 'Size', 'Link'],
                                width=[max(len(a[p]) for a in search_res)
                                       for p in range(0, len(search_res[0]))])

    res = {limit_terminal_size(a): b for a, _, b in search_res}
    result = res[prompt([List('Link', message="",
                              choices=res.keys())])['Link']]
    print(result)

    if opt['--open']:
        return subprocess.check_call(['xdg-open', result])



"""Tets TPB plugin."""


def test_tpb_search_magnets():
    """Test tpb search magnets method."""
    from katcr import ThePirateBay
    from robobrowser import RoboBrowser
    from unittest.mock import patch, MagicMock
    proxies = {'The Pirate Bay': [['foo', None]]}

    with patch('katcr.torrentmirror.get_proxies',
               side_effect=(proxies,)) as mock:
        with patch('katcr.robobrowser.RoboBrowser', spec=RoboBrowser) as mock:
            ThePirateBay(MagicMock()).search_magnets('foo', 1)
            assert mock.open.called_once_with('foo')


def test_tpb_tabulate():
    """Test tabulate method given a known structure.

    Note that this structure may change in the future or even between
    different proxied sites... This needs to be handled somehow.
    """
    from katcr import ThePirateBay
    import bs4
    to_parse = """<div><div class=detName>Foo</div>
                       <div class=detDesc>Bar</div>
                       <a href='foo'></a></div>"""
    fakelink = bs4.BeautifulSoup(to_parse, "html.parser")
    result = ThePirateBay.tabulate(fakelink.find('a'))
    assert result == ('Foo', 'Bar', 'foo')



#!/usr/bin/env python3.5
"""Telegram bot to query KickassTorrents."""

from katcr import Katcr, get_short
from docopt import docopt
from pygogo import Gogo
import telepot
from telepot.loop import MessageLoop
from telepot.namedtuple import InlineKeyboardMarkup, InlineKeyboardButton


class KATBot(telepot.Bot):
    """KAT.cr search bot, looks only for the first page."""

    def __init__(self, opts):
        """Initialize of KATBot."""
        super().__init__(opts['--token'])
        logger = Gogo(__name__, verbose=True)
        self.shortener = opts['--shortener']
        self.katcr = Katcr(logger)
        self.responses = {}

    # pylint: disable=too-few-public-methods
    def on_chat_message(self, msg):
        """Answer only chat messages."""
        if msg['text'] == "/start":
            return

        chat_id = telepot.glance(msg)[2]
        res = tuple(self.katcr.search(msg['text'], 1))

        keyboard = InlineKeyboardMarkup(inline_keyboard=list(
            [InlineKeyboardButton(text=k, callback_data=str(r))]
            for r, (k, _) in enumerate(res)))
        print(keyboard)

        self.responses[chat_id] = {r: (k, v) for r, (k, v) in enumerate(res)}
        self.sendMessage(chat_id, "Results for: {}".format(msg['text']),
                         reply_markup=keyboard, parse_mode="html")

    def on_callback_query(self, msg):
        """Get the button data."""
        _, from_id, query_data = telepot.glance(msg, flavor='callback_query')
        key, value = self.responses[from_id][int(query_data)]
        href = "<a href=\"{}\">{}</a>".format(
            get_short(self.shortener, value), key)
        self.sendMessage(from_id, href, parse_mode="html")


def main():
    """Run telegram bot.

    Usage: katcr_bot [options]

    Options:
        --token=<BOT_TOKEN>             Telegram bot token
        --shortener=<URL_SHORTENER>     Url shortener to use
                                        [default: http://shortmag.net]
    """
    bot = KATBot(docopt(main.__doc__, version="0.0.1"))
    MessageLoop(bot).run_forever()



"""Tests on katcr plugin."""


def test_katcr_search_magnets():
    """Test katcr search magnets method."""
    from katcr import Katcr
    from robobrowser import RoboBrowser
    from unittest.mock import patch, MagicMock
    proxies = {'Kickass Torrents': [['foo', None]]}

    with patch('katcr.torrentmirror.get_proxies',
               side_effect=(proxies,)) as mock:
        with patch('katcr.robobrowser.RoboBrowser', spec=RoboBrowser) as mock:
            Katcr(MagicMock()).search_magnets('foo', 1)
            assert mock.open.called_once_with('foo')


def test_katcr_tabulate():
    """Test tabulate method given a known structure.

    Note that this structure may change in the future or even between
    different proxied sites... This needs to be handled somehow.
    """
    from katcr import Katcr
    import bs4
    to_parse = """<div class=cellMainLink>Foo</div><td>Nope</td><td>Bar</td>
                  <a title='Torrent magnet link' href='foo'></a>"""
    fakelink = bs4.BeautifulSoup(to_parse, "html.parser")
    result = Katcr.tabulate(fakelink)
    assert result == ('Foo', 'Bar', 'foo')



"""Base tests. Nothing related to plugins goes here."""


def test_main():
    """Test argument parsing and calling."""
    from katcr import main, get_from_short
    from unittest.mock import patch, call
    opts = {'<SEARCH_TERM>': "foo", '--search-engine': ['Katcr'],
            '--interactive': False, '--open': False, '-d': False,
            '--disable-shortener': True, '--pages': [1]}
    with patch('katcr.Katcr') as mock:
        with patch('katcr.docopt', side_effect=(opts,)):
            main()
            mock().search.assert_called_with(opts['<SEARCH_TERM>'], 1)

    opts = {'<SEARCH_TERM>': "foo", '--search-engine': ['Katcr'],
            '--interactive': False, '--open': False, '-d': False,
            '--shortener': ['bar'],
            '--disable-shortener': False, '--pages': [1]}
    with patch('katcr.Katcr') as mock:
        with patch('katcr.get_from_short') as short_mock:
            with patch('katcr.docopt', side_effect=(opts,)):
                main()
                mock().search.assert_called_with(opts['<SEARCH_TERM>'], 1)
                short_mock.assert_called_with('bar', [])

    class Foo:
        text = "foo"

    with patch('katcr.requests.post', return_value=Foo) as mock:
        with patch('katcr.docopt', side_effect=(opts,)):
            result = list(get_from_short(
                "foo.com", [("1", "2"), ("3", "4")]))
            assert [result == [('1', '2', 'foo.com/foo'),
                               ('3', '4', 'foo.com/foo')]]
            mock.assert_has_calls([call('foo.com', data={'magnet': '2'}),
                                   call('foo.com', data={'magnet': '4'})])

    opts = {'<SEARCH_TERM>': "foo", '--search-engine': ['Katcr'],
            '--interactive': True, '--open': True, '-d': False,
            '--disable-shortener': True, "--shortener": "http://foo.com",
            '--pages': [1]}
    torr = {'Link': 'foo'}
    args = opts['<SEARCH_TERM>'], 1

    with patch('katcr.Katcr') as mock:
        with patch('katcr.Terminal') as tmock:
            tmock().width = 50
            mock().search.side_effect = ((('foo', "3", 'bar'),),)
            with patch('katcr.subprocess') as smock:
                with patch('katcr.docopt', side_effect=(opts,)):
                    with patch('katcr.prompt', side_effect=(torr,)):
                        main()
                        mock().search.assert_called_with(*args)
                        smock.check_call.assert_called_with(
                            ['xdg-open', 'bar'])

    opts = {'<SEARCH_TERM>': "foo", '--search-engine': ['Katcr'],
            '--interactive': True, '--open': False, '-d': False,
            '--disable-shortener': True, "--shortener": "http://foo.com",
            '--pages': [1]}

    with patch('katcr.Katcr') as mock:
        with patch('katcr.Terminal') as tmock:
            tmock().width = 50
            mock().search.side_effect = ((('foo', "3", 'bar'),),)
            with patch('katcr.subprocess') as smock:
                with patch('katcr.docopt', side_effect=(opts,)):
                    with patch('katcr.prompt', side_effect=(torr,)):
                        main()
                        mock().search.assert_called_with(*args)
                        smock.check_call.assert_not_called()

    opts = {'<SEARCH_TERM>': "foo", '--search-engine': ['Katcr'],
            '--interactive': True, '--open': False,
            '--verbose': True,
            '--disable-shortener': True, '--pages': [1]}

    with patch('katcr.Katcr') as mock:
        with patch('katcr.Terminal') as tmock:
            tmock().width = 50
            mock().search.side_effect = ((('foo', "3", 'bar'),),)
            with patch('katcr.subprocess') as smock:
                with patch('katcr.docopt', side_effect=(opts,)):
                    with patch('katcr.prompt', side_effect=(torr,)):
                        with patch('katcr.Gogo') as logging_mock:
                            main()
                            logging_mock.assert_called_with(
                                'katcr', verbose=True)

    opts = {'<SEARCH_TERM>': "foo",
            '--search-engine': ['Katcr'],
            '--interactive': False,
            '--open': False,
            '--disable-shortener': True,
            "--shortener": "http://foo.com",
            '--pages': [1]}
    torr = {'Link': 'foo'}
    args = opts['<SEARCH_TERM>'], 1

    with patch('katcr.Katcr') as mock:
        tmock().width = 50
        mock().search.side_effect = ((('foo', 'bar', "baz"),),)
        with patch('katcr.docopt', side_effect=(opts,)):
            main()
            mock().search.assert_called_with(*args)


def test_basesearch():
    """Test basesearch has required methods."""
    from katcr import BaseSearch
    import unittest.mock
    import warnings
    warnings.filterwarnings("ignore", category=UserWarning, module='bs4')

    assert hasattr(BaseSearch, "search")
    assert hasattr(BaseSearch, "search_magnets")

    with unittest.mock.patch('katcr.BaseSearch.search_magnets',
                             side_effect=(['foo'],)) as mock:
        class FakeSearch(BaseSearch):
            def get_torrents(self):
                return "foo"
            proxy_name = "The Pirate Bay"
            url = "Foo"
            url_format = None

        FakeSearch(mock.MagicMock).search('foo', 2)
        assert mock.call_count == 2

    class FakeSearch(BaseSearch):
        browser = mock.MagicMock()
        proxy_name = "The Pirate Bay"
        url = "Foo"
        url_format = "http://foo.com/"

        def get_torrents(self):
            return "foo"
    FakeSearch(mock.MagicMock).search('foo', 2)
    assert mock.call_count == 2


def test_cli_help():
    """Test help call."""
    import subprocess
    from katcr import main
    result = subprocess.check_output(['katcr', '--help'])
    assert main.__doc__.encode() in result



"""Bot tests."""


def test_katbot_isinstance():
    """Test katbot is a bot instance."""
    from telepot import Bot
    from katcr.bot import KATBot
    assert issubclass(KATBot, Bot)
    opts = {'--token': "foo", '--shortener': ''}
    assert isinstance(KATBot(opts), Bot)


def test_katbot_main():
    """Test argument parsing and calling."""
    from katcr.bot import main
    from unittest.mock import patch

    opts = {'--token': "foo"}

    with patch('katcr.bot.docopt', side_effect=(opts,)):
        with patch('katcr.bot.KATBot') as bot:
            with patch('katcr.bot.MessageLoop'):
                main()
                bot.assert_called_with({'--token': 'foo'})


def test_on_chat_message_start():
    """Test on chat message handler."""
    from katcr.bot import KATBot

    class FakeKATBot(KATBot):
        """Fake kat bot to avoid initializing telepot."""

        def __init__(self, token):
            """Set token."""
            self.token = token
            self.shortener = "foo"

    fkb = FakeKATBot("foo")
    assert not fkb.on_chat_message({'text': '/start'})


def test_on_chat_message():
    """Test on chat message handler."""
    from katcr.bot import KATBot
    from unittest.mock import patch, MagicMock
    from telepot.namedtuple import InlineKeyboardMarkup, InlineKeyboardButton

    class FakeKATBot(KATBot):
        """Fake kat bot to avoid initializing telepot."""

        def __init__(self, token):
            """Set token."""
            self.token = token
            self.katcr = MagicMock()
            self.katcr.search.return_value = (('foo', 'bar'),)
            self.shortener = "http://foo"
            self.responses = {}
            self.sendMessage = MagicMock()

    with patch('katcr.bot.telepot.glance', return_value=((0, 0, 1))):
        fkb = FakeKATBot("foo")
        assert not fkb.on_chat_message({'text': 'debian'})
        fkb.sendMessage.assert_called_with(
            1, 'Results for: debian',
            parse_mode='html',
            reply_markup=InlineKeyboardMarkup(
                inline_keyboard=[[
                    InlineKeyboardButton(
                        text='foo',
                        url=None, callback_data='0',
                        switch_inline_query=None,
                        switch_inline_query_current_chat=None,
                        callback_game=None, pay=None)]]))
        assert fkb.responses


def test_on_chat_message_empty():
    """Test on chat message handler."""
    from katcr.bot import KATBot
    from unittest.mock import patch, MagicMock
    from telepot.namedtuple import InlineKeyboardMarkup

    class FakeKATBot(KATBot):
        """Fake kat bot to avoid initializing telepot."""

        def __init__(self, token):
            """Set token."""
            self.token = token
            self.katcr = MagicMock()
            self.shortener = "http://foo"
            self.responses = {}
            self.sendMessage = MagicMock()

    with patch('katcr.bot.telepot.glance', return_value=((0, 0, 1))):
        fkb = FakeKATBot("foo")
        assert not fkb.on_chat_message({'text': 'debian'})
        fkb.sendMessage.assert_called_with(
            1, 'Results for: debian',
            parse_mode='html',
            reply_markup=InlineKeyboardMarkup(inline_keyboard=[]))


def test_on_callback_query():
    """Test on chat message handler."""
    from katcr.bot import KATBot
    from unittest.mock import patch, MagicMock

    class FakeKATBot(KATBot):
        """Fake kat bot to avoid initializing telepot."""

        def __init__(self, token):
            """Set token."""
            self.token = token
            self.katcr = MagicMock()
            self.responses = {1: {1: ['foo', 'foo']}}
            self.shortener = "http://foo"
            self.sendMessage = MagicMock()

    with patch('katcr.bot.telepot.glance', return_value=((0, 1, 1))):
        with patch('katcr.bot.get_short', return_value=(('foo'))):
            fkb = FakeKATBot("foo")
            assert not fkb.on_callback_query({'text': 'debian'})
            assert fkb.sendMessage.called
            assert fkb.sendMessage.call_args[0][0] == 1
            assert 'href' in fkb.sendMessage.call_args[0][1]
            assert fkb.sendMessage.call_args[1]['parse_mode'] == 'html'
